#include <Columns/ColumnString.h>
#include <Columns/ColumnVector.h>
#include <DataTypes/DataTypesNumber.h>
#include <Disks/IDisk.h>
#include <Functions/FunctionFactory.h>
#include <Functions/IFunction.h>
#include <Interpreters/Context.h>
#include <Poco/Util/AbstractConfiguration.h>

namespace DB
{
namespace ErrorCodes
{
    extern const int ILLEGAL_COLUMN;
    extern const int ILLEGAL_TYPE_OF_ARGUMENT;
    extern const int NUMBER_OF_ARGUMENTS_DOESNT_MATCH;
    extern const int UNKNOWN_DISK;
}
namespace
{

struct FilesystemAvailable
{
    static constexpr auto name = "filesystemAvailable";
    static std::uintmax_t get(const DiskPtr & disk) { return disk->getAvailableSpace(); }
};

struct FilesystemUnreserved
{
    static constexpr auto name = "filesystemUnreserved";
    static std::uintmax_t get(const DiskPtr & disk) { return disk->getUnreservedSpace(); }
};

struct FilesystemCapacity
{
    static constexpr auto name = "filesystemCapacity";
    static std::uintmax_t get(const DiskPtr & disk) { return disk->getTotalSpace(); }
};

template <typename Impl>
class FilesystemImpl : public IFunction
{
public:
    static constexpr auto name = Impl::name;

    static FunctionPtr create(ContextPtr context_) { return std::make_shared<FilesystemImpl<Impl>>(context_); }

    explicit FilesystemImpl(ContextPtr context_) : context(context_) { }

    bool useDefaultImplementationForConstants() const override { return true; }

    bool isSuitableForShortCircuitArgumentsExecution(const DataTypesWithConstInfo & /*arguments*/) const override
    {
        return false;
    }

    String getName() const override { return name; }

    bool isVariadic() const override { return true; }

    size_t getNumberOfArguments() const override { return 0; }
    bool isDeterministic() const override { return false; }

    DataTypePtr getReturnTypeImpl(const DataTypes & arguments) const override
    {
        if (arguments.size() > 1)
        {
            throw Exception("Arguments size of function " + getName() + " should be 0 or 1", ErrorCodes::NUMBER_OF_ARGUMENTS_DOESNT_MATCH);
        }
        if (arguments.size() == 1 && !isStringOrFixedString(arguments[0]))
        {
            throw Exception(
                "Arguments of function " + getName() + " should be String or FixedString", ErrorCodes::ILLEGAL_TYPE_OF_ARGUMENT);
        }
        return std::make_shared<DataTypeUInt64>();
    }

    ColumnPtr executeImpl(const ColumnsWithTypeAndName & arguments, const DataTypePtr &, size_t input_rows_count) const override
    {
        if (arguments.empty())
        {
            auto disk = context->getDisk("default");
            return DataTypeUInt64().createColumnConst(input_rows_count, Impl::get(disk));
        }
        else
        {
            auto col = arguments[0].column;
            if (const ColumnString * col_str = checkAndGetColumn<ColumnString>(col.get()))
            {
                auto disk_map = context->getDisksMap();

                auto col_res = ColumnVector<UInt64>::create(col_str->size());
                auto & data = col_res->getData();
                for (size_t i = 0; i < col_str->size(); ++i)
                {
                    auto disk_name = col_str->getDataAt(i).toString();
                    if (auto it = disk_map.find(disk_name); it != disk_map.end())
                        data[i] = Impl::get(it->second);
                    else
                        throw Exception(
                            "Unknown disk name " + disk_name + " while execute function " + getName(), ErrorCodes::UNKNOWN_DISK);
                }
                return col_res;
            }
            throw Exception(
                "Illegal column " + arguments[0].column->getName() + " of argument of function " + getName(), ErrorCodes::ILLEGAL_COLUMN);
        }
    }

private:
    ContextPtr context;
};

}

REGISTER_FUNCTION(Filesystem)
{
    factory.registerFunction<FilesystemImpl<FilesystemAvailable>>();
    factory.registerFunction<FilesystemImpl<FilesystemCapacity>>();
    factory.registerFunction<FilesystemImpl<FilesystemUnreserved>>();
}

}
